/*
 * XING header manipulation routines
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */


#ifndef lint
static char SccsId[] = "$Id: xing_header.c,v 1.7.6.1 2003/02/06 17:34:35 number6 Exp $";
#endif

#include <stdio.h>
#include "xing_header.h"

#if 0
#define UNIT_TEST 1
#endif

static int sr_table[4] = { 44100, 48000, 32000, 99999 };

/*  big endian insert */

void InsertI4(int x, unsigned char *buf)
{
    buf[3] = x & 0xff;
    x >>= 8;

    buf[2] = x & 0xff;
    x >>= 8;

    buf[1] = x & 0xff;
    x >>= 8;

    buf[0] = x & 0xff;
}

/*  big endian extract */

int ExtractI4(unsigned char *buf)
{
    int x;

    x = buf[0];
    x <<= 8;
    x |= buf[1];
    x <<= 8;
    x |= buf[2];
    x <<= 8;
    x |= buf[3];

    return x;
}


void xingheader_edit(XHEADDATA *xingh, int frames, int bytes, int id, int mode)
{
    edit_xing_header(frames, bytes, xingh->xingbuf, id, mode);
}


int xingheader_parse(XHEADDATA *xingh)
{
    int status;

    status = get_xing_header(xingh->xingbuf, xingh);
    return status;
}


void xingheader_print(XHEADDATA *xingh)
{
    print_xing_header(xingh);
}


/*
 * Poke new values for frames and file size into an already populated
 * and valid Xing header. 'version' is the raw mpeg version value
 * from the decoded header. 'mode' is the raw mpeg channel mode value
 * from the decoded header.
 */
void edit_xing_header(int frames, int bytes, 
                      unsigned char *buf, int version, int mode)
{
    int offset = 44;

    switch (version) {
      case 3:                /* mpeg 1 */
        switch (mode) {
          case 3:            /* mono */
            offset = 29;
            break;
          default:
            offset = 44;     /* stereo/joint/dual channel */
            break;
        }
        break;
      default:               /* mpeg 2/2.5 */
        switch (mode) {
          case 3:            /* mono */
            offset = 21;
            break;
          default:           /* stereo/joint/dual channel */
            offset = 29;
            break;
        }
        break;
    }
    InsertI4(frames, &buf[offset]);
    InsertI4(bytes,  &buf[offset+4]);
}


/* Return TRUE if valid Xing header is found, FALSE otherwise */

int get_xing_header(unsigned char *buf, XHEADDATA *X)
{
    int h_id;
    int h_id2;
    int h_layer;
    int h_protect;
    int h_sr_index;
    int h_mode;
    int head_flags;
    int i;

    /* 11111111 111BBCCD EEEEFFGH IIJJKLMM  */

    /* BB */
    h_id       = (buf[1] >> 3) & 1;
    h_id2      = (buf[1] >> 3) & 2;

    /* CC */
    h_layer    = (buf[1] >> 1) & 3;

    /* D */
    h_protect  =  buf[1]       & 1;

    h_sr_index = (buf[2] >> 2) & 3;
    h_mode     = (buf[3] >> 6) & 3;

    if (h_id) {        /*  mpeg1 */
        if (h_mode != 3) {
            buf += (32+4);
        }
        else {
            buf += (17+4);
        }
    }
    else {      /* mpeg2 */
        if (h_mode != 3) {
            buf += (17+4);
        }
        else {
            buf += (9+4);
        }
    }

    if (h_layer == 3) {
        X->h_layer = 1;
    }
    else if (h_layer == 2) {
        X->h_layer = 2;
    }
    else if (h_layer == 1) {
        X->h_layer = 3;
    }
    else {
        X->h_layer = 0;
    }

    X->h_protect = !h_protect;

    /* fail, header not found */

    if (buf[0] != 'X') return 0;
    if (buf[1] != 'i') return 0;
    if (buf[2] != 'n') return 0;
    if (buf[3] != 'g') return 0;
    buf+=4;

    X->h_id = h_id2;
    X->samprate = sr_table[h_sr_index];
    if (h_id == 0) {
        X->samprate >>= 1;
        if (h_id2 == 0) X->samprate >>= 1;
    }

    /* get flags */

    head_flags = X->flags = ExtractI4(buf);
    buf+=4;

    if (head_flags & FRAMES_FLAG) {
        X->frames  = ExtractI4(buf);
        buf+=4;
    }
    if (head_flags & BYTES_FLAG) {
        X->bytes = ExtractI4(buf);
        buf+=4;
    }

    if (head_flags & TOC_FLAG) {
        if (X->toc != NULL) {
            for(i=0;i<100;i++) {
                X->toc[i] = buf[i];
            }
        }
        buf+=100;
    }

    X->vbr_scale = -1;
    if (head_flags & VBR_SCALE_FLAG)  {
       X->vbr_scale = ExtractI4(buf); buf+=4;
    }
    return 1;
}


void print_xing_header(XHEADDATA *X)
{
    printf("h_id      = %d\n", X->h_id);
    printf("h_layer   = %d\n", X->h_layer);
    printf("h_protect = %d\n", X->h_protect);
    printf("samprate  = %d\n", X->samprate);
    printf("flags     = %d\n", X->flags);
    printf("frames    = %d\n", X->frames);
    printf("bytes     = %d\n", X->bytes);
    printf("vbr_scale = %d\n", X->vbr_scale);
}


#if defined(UNIT_TEST)
int main(int argc, char *argv[])
{
    unsigned char buf[XING_HEADER_SIZE];
    FILE *fp;
    int items;
    XHEADDATA X;
    int i;
    int frame_add = 0;
    int bytes_add = 0;
    int frames = 0;
    int bytes = 0;
    char *filename;
    char *cp;

    memset(buf, 0, sizeof(buf));
    InsertI4(0x010f03f0, buf); 

    memset(&X, 0, sizeof(X));

    if (argc == 1 || argc > 4) {
        printf("usage: %s mp3_file [[+]frames] [[+]bytes]\n", argv[0]);
        return 1;
    }
    filename = argv[1];

    if (argc > 2) {
        cp = argv[2];
        if (*cp == '+') {
            frame_add = 1;
            cp++;
        }
        frames = atoi(cp);
    }
    if (argc == 4) {
        cp = argv[3];
        if (*cp == '+') {
            bytes_add = 1;
            cp++;
        }
        bytes = atoi(cp);
    }

    fp = fopen(filename, "rb+");
    if (!fp) {
        perror("fopen");
        return 1;
    }
    items = fread(buf, XING_HEADER_SIZE, 1, fp);

    for (i = 0; i < 64; i++) {
        if ((i%16) == 0) {
            printf("\n");
        }
        printf("%02x ", buf[i]);
    }
    printf("\n");
    get_xing_header(buf, &X);

    printf("h_id      = %d\n", X.h_id);
    printf("h_layer   = %d\n", X.h_layer);
    printf("h_protect = %d\n", X.h_protect);
    printf("samprate  = %d\n", X.samprate);
    printf("flags     = %d\n", X.flags);
    printf("frames    = %d\n", X.frames);
    printf("bytes     = %d\n", X.bytes);
    printf("vbr_scale = %d\n", X.vbr_scale);
    printf("==================================\n");

    if (frame_add) {
        frames += X.frames;
    }
    if (bytes_add) {
        bytes += X.bytes;
    }
    if (frames) {
        if (bytes == 0) {
            bytes = X.bytes;
        }
        edit_xing_header(frames, bytes, buf);
        get_xing_header(buf, &X);
        printf("h_id      = %d\n", X.h_id);
        printf("h_layer   = %d\n", X.h_layer);
        printf("h_protect = %d\n", X.h_protect);
        printf("samprate  = %d\n", X.samprate);
        printf("flags     = %d\n", X.flags);
        printf("frames    = %d\n", X.frames);
        printf("bytes     = %d\n", X.bytes);
        printf("vbr_scale = %d\n", X.vbr_scale);
        printf("==================================\n");
        fseek(fp, 0, SEEK_SET);
        fwrite(buf, XING_HEADER_SIZE, 1, fp);
        fclose(fp);
    }
    return 0;
}
#endif
